/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "tree_constructor.h"
#include    <set>

#ifdef      LTREE_DEBUGGING
#include    "tree_constructor_debug.cpp"
#endif

namespace tree{

//-----------------------------------------------------------------------------
// Expand branch drafts, return a vector of expanded branch drafts and an error
// message (empty if there are no errors)

pair<vector<TBranchDraft>, string> ExpandBranchDrafts(
        const vector<TBranchDraft> &drafts){

    #ifdef LTREE_DEBUGGING
    cout << "Expanding branch drafts..." << endl;
    for(const auto &draft : drafts){
        cout << "\t";
        draft.Print();
        }
    #endif

    vector<TBranchDraft> res;

    // Adding an empty branch draft if there is no branch drafts at all
    // in order to take a root node into account
    if(drafts.empty()){
        #ifdef LTREE_DEBUGGING
        cout << "\t- There are no branch drafts" << endl;
        #endif
        TBranchDraft draft;
        draft.Complete(0);
        res.push_back(draft);
        return {res, ""};
        }

    string prev_node_name;
    set<int> branch_marks({-1});
    set<int> master_branches;

    // For all initial branch drafts
    for(const auto &draft : drafts){

        branch_marks.insert(draft.BranchMark());
        master_branches.insert(draft.MasterBranch());

        vector<TBranchDraft> xdrafts(1);

        // For all nodes of an initial branch draft
        for(const auto &node_name : draft.Nodes()){

            // If this node is not optional
            if(node_name.empty() || node_name.front() != '?'){
                // Adding this node to previously expanded nodes
                for(auto &xdraft : xdrafts){
                    xdraft.Add({node_name});
                    }
                prev_node_name = node_name;
                continue;
                }

            if(node_name.size() < 2){
                return {res, "Node names consisting of a single '?' character "
                    "are not allowed"};
                }

            const auto node_name_prep = node_name.substr(1);

            // Expanding this optional node
            if(prev_node_name != node_name){
                const size_t l = xdrafts.size();
                for(size_t i = 0; i < l; i++){
                    xdrafts.push_back(xdrafts.at(i));
                    }
                for(size_t i = l ; i < l * 2; i++){
                    xdrafts.at(i).Add({node_name_prep});
                    }
                }
            else{
                xdrafts.push_back(xdrafts.back());
                xdrafts.back().Add({node_name_prep});
                }

            prev_node_name = node_name;
            }

        // Completing expanded branch drafts
        for(auto &xdraft : xdrafts){
            if(!draft.Attr().empty()){
                xdraft.Add({}, draft.Attr());
                }
            xdraft.Complete(draft.BranchMark(), draft.MasterBranch(),
                draft.IndexEnabled(), draft.IsAnchorBranch());
            }

        // Inserting expanded branch drafts into resulted vector
        res.insert(res.end(), xdrafts.begin(), xdrafts.end());
        }

    // Checking that every master branch is marked with an existing branch mark
    for(const auto &mb : master_branches){
        if(branch_marks.find(mb) == branch_marks.end()){
            return {{}, "There is no branch mark '" + to_string(mb) + "'"};
            }
        }

    #ifdef LTREE_DEBUGGING
    cout << "Branch drafts are expanded:" << endl;
    for(const auto &draft : res){
        cout << "\t";
        draft.Print();
        }
    #endif

    return {res, ""};
    }

//-----------------------------------------------------------------------------
// Check whether a given node exists in the tree

bool TTreeConstructor::IsPresent(TNode &node) const{

    for(const auto &el : nodes){
        if(el == node){
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Find node by specified name and parent ID

TNode* TTreeConstructor::FindNode(const string &node_name, const int parent_id){

    for(auto &n : nodes){
        if(n.Name() == node_name && n.ParentId() == parent_id){
            return &n;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Find node by specified ID

TNode* TTreeConstructor::FindNode(const int id){

    for(auto &node : nodes){
        if(node.Id() == id){
            return &node;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Find node that has specified name and that exists on a specified level

TNode* TTreeConstructor::FindNodeOnLevel(const string &node_name,
        const int level){

    map<int, vector<TNode*>> leveled_nodes;

    // Get pointers to nodes that stand on a same level
    for(auto &node : nodes){
        leveled_nodes[GetLevel(node)].push_back(&node);
        }

    // Search through the leveled nodes
    if(leveled_nodes.find(level) != leveled_nodes.end()){
        for(auto node : leveled_nodes[level]){
            if(node && node->Name() == node_name){
                return node;
                }
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Get a level at which a given node is located

int TTreeConstructor::GetLevel(const TNode &node){

    int parent_id = node.ParentId();
    int level = 0;

    while(parent_id >= 0){
        level++;
        TNode *n = FindNode(parent_id);
        if(!n){
            return -1;
            }
        parent_id = n->ParentId();
        }

    return level;
    }

//-----------------------------------------------------------------------------
// Add a new node as a child of a given parent or as a child of a currently
// selected node

TNode* TTreeConstructor::NewNode(const string &name, TNode *parent){

    if(!parent){
        if(nodes.empty()){
            return NULL;
            }
        parent = &nodes.front();
        }

    // If given parent does not exist in this tree
    if(!IsPresent(*parent)){
        return NULL;
        }

    // Not duplicating the nodes
    TNode *existing_node = FindNode(name, parent->Id());
    if(existing_node){
        return existing_node;
        }

    // Generating a unique ID
    int unique_id = 0;
    for(const auto &node : nodes){
        if(node.Id() >= unique_id){
            unique_id = node.Id() + 1;
            }
        }

    // Creating a new node that will have specified name, unique ID and
    // a specified parent ID
    TNode new_node(name, unique_id, parent->Id());

    #ifdef LTREE_DEBUGGING
    cout << "\t\t-- Pushing new node " << new_node.Name() << endl;
    #endif

    // Saving this new node
    nodes.push_back(new_node);

    // Updating a vector of childs of a parent of this new node
    if(nodes.back().ParentId() >= 0){
        TNode *new_parent = FindNode(nodes.back().ParentId());
        if(!new_parent){
            return NULL;
            }
        new_parent->AddChild(nodes.back().Id());
        }

    return &nodes.back();
    }

//-----------------------------------------------------------------------------
// Add a new end-text-node to this tree as a child of a given parent

TNode* TTreeConstructor::EndNode(const string &name, const int branch_mark,
        const int master_branch, const bool index_enabled,
        const bool is_anchor_branch, TNode *parent){

    if(branch_mark < 0){
        return NULL;
        }

    if(!parent){
        TNode *root = FindRoot();
        NewBranch(root, branch_mark, master_branch, index_enabled,
            is_anchor_branch);
        return root;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\t\t-- Preparing new end node "
         << (name.empty() ? EMPTY_NODE : "\"" + name + "\"") << endl;
    #endif

    TNode *current = NewNode(name, parent);

    if(current && !is_anchor_branch && doc_type == TYPE_XML){
        current = NewNode("text", current);
        }
    if(current){
        NewBranch(current, branch_mark, master_branch, index_enabled,
            is_anchor_branch);
        }

    return current;
    }

//-----------------------------------------------------------------------------
// Add a new attribute to a given node

bool TTreeConstructor::NewAttribute(const string &attr, TNode *node){

    if(!node || !IsPresent(*node)){
        return false;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\t\t-- Saving new attrubute '" << attr << "'" << endl;
    #endif

    node->WriteAttribute(attr);

    return true;
    }

//-----------------------------------------------------------------------------
// Create a new branch from all ancestors of a given "leaf" node + this "leaf"
// node itself, mark this branch with a given branch mark and, optionaly, make
// this branch an additional branch of a specified master branch

bool TTreeConstructor::NewBranch(TNode *leaf, const int branch_mark,
        const int master_branch, const bool index_enabled,
        const bool is_anchor_branch){

    if(!leaf || !IsPresent(*leaf)){
        return false;
        }

    TBranch branch;
    branch.push_back(*leaf);

    // Finding ancestors of a last element of a new branch and adding them in
    if(!TraceParents(leaf, branch)){
        cout << "[ERROR: Parent node is missing]" << endl;
        }

    reverse(branch.begin(), branch.end());
    branch.WriteBranchMark(branch_mark);
    branch.WriteMasterBranch(master_branch);
    branch.WriteIndexFlag(index_enabled);
    branch.WriteAnchorFlag(is_anchor_branch);

    #ifdef LTREE_DEBUGGING
    cout << "\t\t-- Pushing new branch..." << endl << "\t";
    branch.Print();
    #endif

    branches.push_back(branch);

    return true;
    }

//-----------------------------------------------------------------------------
// Trace all parents of a given node and push them into the "parents" vector

bool TTreeConstructor::TraceParents(TNode *node, vector<TNode> &parents){

    if(!node){
        return true;
        }

    TNode *parent;
    TNode *current_node = node;

    for(;;){
        if(current_node->ParentId() < 0){
            break;
            }
        parent = FindNode(current_node->ParentId());
        if(!parent){
            return false;
            }
        // Adding a parent
        parents.push_back(*parent);
        current_node = parent;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Construct a tree using given base nodes and a branch draft

pair<pair<vector<TNode>, vector<TBranch>>, bool> TTreeConstructor::Construct(
        const string &root_name, const vector<TBranchDraft> &branch_drafts,
        const int dt){

    doc_type = dt;

    #ifdef LTREE_DEBUGGING
    cout << "Starting construction of the tree" << endl;
    #endif

    // Writing parameters of a root node
    TNode root(root_name, 0, -1);

    #ifdef LTREE_DEBUGGING
    cout << "Root node is "
         << (root_name.empty() ? EMPTY_NODE : "\"" + root_name + "\"")
         << endl;
    #endif

    nodes.clear();
    nodes.push_back(root);
    branches.clear();

    // Expanding branch drafts
    const auto [expanded_branch_drafts, err] = ExpandBranchDrafts(branch_drafts);
    if(!err.empty()){
        cout << "[ERROR: " << err << "]" << endl;
        return {{}, false};
        }

    // For all expanded branch drafts
    for(const auto &branch_draft : expanded_branch_drafts){

        #ifdef LTREE_DEBUGGING
        branch_draft.Print();
        #endif

        const auto &draft_nodes  = branch_draft.Nodes();
        const auto branch_mark   = branch_draft.BranchMark();
        const auto master_branch = branch_draft.MasterBranch();
        const auto index_enabled = branch_draft.IndexEnabled();
        const auto is_anchor_branch = branch_draft.IsAnchorBranch();

        // Handling empty branch drafts
        if(draft_nodes.empty()){
            if(!EndNode("", branch_mark, master_branch, index_enabled,
                    is_anchor_branch, NULL)){
                return {{}, false};
                }
            continue;
            }

        TNode *current = FindRoot();
        bool original_part_of_the_branch_draft = false;

        // For all nodes of a branch draft
        for(auto draft_node = draft_nodes.begin();
                draft_node != draft_nodes.end();
                draft_node++){

            TNode *present = NULL;

            // No need to check if a node is present in the tree
            // if it's parent is not present in the tree
            if(!original_part_of_the_branch_draft){
                // Get node with same name and level
                const int level = draft_node - draft_nodes.begin() + 1;
                present = FindNodeOnLevel(*draft_node, level);
                if(!present){
                    original_part_of_the_branch_draft = true;
                    }
                #ifdef LTREE_DEBUGGING
                cout << "\t- "
                     << (draft_node->empty() ? EMPTY_NODE : *draft_node)
                     << " is " << (present ? "" : "not ") << "present;" << endl;
                #endif
                }

            // Adding a final node of a branch draft
            if(draft_node == (draft_nodes.end() - 1)){
                // If node has an attribute
                if(!branch_draft.Attr().empty()){
                    current = NewNode(*draft_node, current);
                    if(!current ||
                            !NewAttribute(branch_draft.Attr(), current) ||
                            !NewBranch(current, branch_mark, master_branch,
                                index_enabled, is_anchor_branch)){
                        return {{}, false};
                        }
                    }
                else{
                    if(!EndNode(*draft_node, branch_mark, master_branch,
                            index_enabled, is_anchor_branch, current)){
                        return {{}, false};
                        }
                    }
                current = FindRoot();
                }

            // Adding a node that is not a final node of this branch draft
            else if(!present){
                current = NewNode(*draft_node, current);
                if(!current){
                    return {{}, false};
                    }
                }

            if(present){
                current = present;
                }
            }
        }

    #ifdef LTREE_DEBUGGING
    PrintTree();
    PrintBranches();
    #endif

    return {{nodes, branches}, true};
    }

//-----------------------------------------------------------------------------

} // namespace tree
